
#include <QHBoxLayout>
#include <QLabel>
#include <QDebug>
#include <QDialogButtonBox>
#include <QMessageBox>
#include <QPushButton>
#include <QLibraryInfo>
#include <qwinwidget.h>

#include <qt_windows.h>

#ifdef QT_DIAG_LIB
# include <nativewindowdump.h>
#endif

HWND createdHwnd = 0;
HWND mainWinWidgetHwnd = 0;

QString description()
{
#if QT_VERSION >= 0x050000
    return QLibraryInfo::build();
#else
    return QT_VERSION_STR;
#endif
}

static void dumpAncestors(QDebug d, HWND hwnd)
{

    LONG_PTR gwlP = GetWindowLongPtr(hwnd, -8 /* GWL_HWNDPARENT */);
    d << "Ancestors of" << hwnd
      << "GWL_HWNDPARENT=" << reinterpret_cast<HWND>(gwlP)
      << "GA_PARENT=" <<  GetAncestor(hwnd, GA_PARENT)
      << "GA_ROOT=" << GetAncestor(hwnd, GA_ROOT)
      << "GA_ROOTOWNER=" << GetAncestor(hwnd, GA_ROOTOWNER);

}

class MyMessageBox  : public QDialog {
    Q_OBJECT
public:
    MyMessageBox(QWidget *p);
    void setVisible(bool v)
    {
        QDialog::setVisible(v);
        QDebug d = qDebug();
        d << __FUNCTION__ << v;
        if (v) {
            if (WId wid = internalWinId()) {
                HWND hwnd = reinterpret_cast<HWND>(wid);
                d << "internalWinId=" << hwnd;
                dumpAncestors(d, hwnd);
            }
        }
    }

public slots:
    void dump()
    {
#ifdef QT_DIAG_LIB
        qDebug() << ">Dump starting at " << createdHwnd;
        QtDiag::dumpNativeWindows(WId(createdHwnd));
        qDebug("<Dump");
#endif
    }
    void dumpAll()
    {
#ifdef QT_DIAG_LIB
        qDebug(">Full Dump");
        QtDiag::dumpNativeWindows();
        qDebug("<Full Dump");
#endif
    }
};

MyMessageBox::MyMessageBox(QWidget *p) : QDialog (p)
{
    setWindowTitle(QLatin1String("Dialog ") + QLatin1String(QT_VERSION_STR));
    QVBoxLayout *lt = new QVBoxLayout(this);
    lt->addWidget(new QLabel(description()));

    QDialogButtonBox *bb = new QDialogButtonBox(QDialogButtonBox::Close, Qt::Horizontal, this);
    connect(bb, SIGNAL(rejected()), this, SLOT(reject()));
    lt->addWidget(bb);
    QPushButton *db = bb->addButton("Dump", QDialogButtonBox::ActionRole);
    connect(db, SIGNAL(clicked()), this, SLOT(dump()));
    db = bb->addButton("Dump All", QDialogButtonBox::ActionRole);
    connect(db, SIGNAL(clicked()), this, SLOT(dumpAll()));
}

LRESULT CALLBACK WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    switch (message)  {
    case WM_LBUTTONUP:	{
        qDebug("WM_LBUTTONUP");
        QWinWidget dialogW(hWnd, 0, 0);        
        dialogW.showCentered();
        const HWND dhwnd = reinterpret_cast<HWND>(dialogW.winId());
        SetWindowText(dhwnd, L"DialogWinWidget");
        {
            QDebug d = qDebug();
            d << "QWinWidget(dialogW)=" << dhwnd;
            dumpAncestors(d, dhwnd);
        }
        MyMessageBox mb(&dialogW);
        qDebug() << ">QMessageBox::exec()";
        int result = mb.exec();
        qDebug() << "<QMessageBox::exec()" << result;
    }
        break;

    case WM_KEYDOWN:
        qDebug() << "WM_KEYDOWN" << wParam;
        switch (wParam) {
        case VK_TAB:
            qDebug() << "SetFocus" << mainWinWidgetHwnd;
            SetFocus(mainWinWidgetHwnd);
            break;
        case VK_F1:
#ifdef QT_DIAG_LIB
            QtDiag::dumpNativeWindows(WId(hWnd));
#endif
            break;
        default:
            return DefWindowProc(hWnd, message, wParam, lParam);
        }
        break;

    case WM_SETFOCUS:
        qDebug() << "WM_SETFOCUS" << reinterpret_cast<HWND>(wParam);
        break;

    case WM_KILLFOCUS:
        qDebug() << "WM_KILLFOCUS" << reinterpret_cast<HWND>(wParam);
        break;

    case WM_DESTROY:
        qDebug("WM_DESTROY");
        PostQuitMessage(0);
        break;

    case WM_ACTIVATE:
        qDebug() << "WM_ACTIVATE wParam: high(minState)=" << HIWORD(wParam)
                 << "low(active)=" << LOWORD(wParam) << "lParam(HWND)="
                 << reinterpret_cast<HWND>(lParam);
        break;

    case WM_ACTIVATEAPP:
        qDebug() << "WM_ACTIVATEAPP wParam (active)=" << wParam << "lParam(thread)=" << lParam;
        break;

    default:
        return DefWindowProc(hWnd, message, wParam, lParam);
    }
    return 0;
}

#ifdef USE_WINMAIN
int APIENTRY wWinMain(HINSTANCE hInstance, HINSTANCE /*hPrevInstance*/, LPTSTR    /*lpCmdLine*/, int       nCmdShow)
{
#else
int main(int, char **)
{
    HINSTANCE hInstance = static_cast<HINSTANCE>(GetModuleHandle(0));
    const int nCmdShow = SW_SHOW;
#endif
    WNDCLASSEX wcex;
    
    wcex.cbSize = sizeof(WNDCLASSEX); 
    
    wcex.style		= CS_HREDRAW | CS_VREDRAW;
    wcex.lpfnWndProc	= (WNDPROC)WndProc;
    wcex.cbClsExtra	= 0;
    wcex.cbWndExtra	= 0;
    wcex.hInstance	= hInstance;
    wcex.hIcon		= NULL;
    wcex.hCursor	= LoadCursor(NULL, IDC_ARROW);
    wcex.hbrBackground	= (HBRUSH)(COLOR_WINDOW+1);
    wcex.lpszMenuName	= NULL;
    wcex.lpszClassName	= L"qtest";
    wcex.hIconSm	= NULL;

    ATOM windowClass = RegisterClassEx(&wcex);

    QString t = description();
    createdHwnd = CreateWindow(reinterpret_cast<TCHAR*>(windowClass),
                               reinterpret_cast<const TCHAR*>(t.utf16()),
                               WS_OVERLAPPEDWINDOW, CW_USEDEFAULT, 0, CW_USEDEFAULT, 0, 0, 0, hInstance, 0);

    qDebug() << t << '\n' << "(created) hwnd=" << createdHwnd;
    if (!createdHwnd)
        return FALSE;

    int argc = 0;
    QApplication a(argc, 0);

    QWinWidget win(createdHwnd);
    mainWinWidgetHwnd = reinterpret_cast<HWND>(win.winId());
    SetWindowText(mainWinWidgetHwnd, L"MainWinWidget");
    qDebug() << "QWinWidget() winId=" << mainWinWidgetHwnd;

    QHBoxLayout *lt = new QHBoxLayout(&win);
    lt->addWidget(new QLabel(t));
    
    win.move(0, 0);
    win.show();

    ShowWindow(createdHwnd, nCmdShow);
    UpdateWindow(createdHwnd);

    return a.exec();
}

#include "main.moc"
